#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#include "tidedb.h"
#include "PDB.h"

#define	kHighTide	1
#define	kLowTide	0

#define	kLineSize	1024

#define	kHeaderDelims	"\t\n\r"
#define kFieldDelims	" ,\t\n\r"

#if !defined (false)
enum {false = 0, true};
#endif

void main(void);
char *getnext(FILE *fp, int nextLine, char *delims);
void convert(FILE *fs);
int convert_hltable(FILE *fs, FILE *fd, TideDatabaseHeader *header, int format);
int convert_xtide(FILE *fs, FILE *fd, TideDatabaseHeader *header, int format);
char *strlower(char *p);
void setshort(unsigned short *s, unsigned short value);
void setlong(unsigned long *l, unsigned long value);


void main(void)
{
	FILE	*fs;
	char	filename[80];
	
RESTART:
	printf("Tide filename to convert: ");
	if (fgets(filename,80,stdin) == NULL)
		return;
		
	sscanf(filename, "%s\n", filename);
	
	fs = fopen(filename, "r");
	if (fs == NULL)
	{
		printf("ERROR: Couldn't open %s\n", filename);
		goto RESTART;
	}
	convert(fs);
	
	fclose(fs);
	
	printf("File converted.\n\n");
	
	printf("Press Return/Enter to exit.\n");
	fgets(filename,80,stdin);
}


char *getnext(FILE *fp, int nextLine, char *delims)
{
	static char	line[kLineSize];
	static int	linecount = 0;
	char		*token;
	
NEXT:
	if (nextLine)
	{
		linecount++;
		token = fgets(line, kLineSize, fp);
		
		if (token != NULL)
		{
			if (delims != kHeaderDelims)
				strlower(line);
			token = strtok(line, delims);
			if (*token == '#')
				goto NEXT;
		}
	}	
	else
		token = strtok(NULL, delims);

	return(token);
}

void convert(FILE *fs)
{
	FILE				*fd;
	int					err;
	PBDHeader			pdbHeader;
	PDBRecordEntry		pdbEntry;		
	TideDatabaseHeader	header;
	char				*dosName,*pdbName,*siteName;
	char				*token;
	int					format;
	int					tidesPerDay;
	time_t				now;
	
	memset(&pdbHeader, 0, kPDBHeaderSize);
	memset(&pdbEntry, 0, kPDBRecordEntrySize);
	memset(&header, 0, kTideDatabaseHeaderSize);

	dosName = getnext(fs,true,kHeaderDelims);
	if (dosName == NULL)
	{
		fprintf(stderr, "Error: File format error -- could not read DOS filename\n");
		goto ERROR;
	}
	
	pdbName = getnext(fs,false,kHeaderDelims);
	if (pdbName == NULL)
	{
		fprintf(stderr, "Error: File format error -- could not read PDB filename\n");
		goto ERROR;
	}

	siteName = getnext(fs,false,kHeaderDelims);
	if (siteName == NULL)	
	{
		fprintf(stderr, "Error: File format error -- could not read site name\n");
		goto ERROR;
	}
	
	fd = fopen(dosName,"wb");
	if (fd == NULL)
	{
		fprintf(stderr, "Error: could not create %s\n", dosName);
		goto ERROR;
	}

	time(&now);
	strcpy(pdbHeader.name, pdbName);
	setshort(&(pdbHeader.flags), pdbBackupFlag);
	setlong(&(pdbHeader.type), 'Tdat');
	setlong(&(pdbHeader.creator), 'Tide');
	setlong(&(pdbHeader.creationTime), now);
	setlong(&(pdbHeader.modificationTime), now);
	setshort(&(pdbHeader.numRecords), 1);
	fwrite(&pdbHeader,kPDBHeaderSize,1,fd);

	setlong(&(pdbEntry.offset), kPDBFirstRecordOffset);
	fwrite(&pdbEntry,kPDBRecordEntrySize,1,fd);
	fputc(0, fd);
	fputc(0, fd);

	strcpy(header.site, siteName);
	header.version = kTidesDatabaseVersion;
	
	//	Get the file format, units, and tidesPerDay
	token = getnext(fs,true,kFieldDelims);
	if (token == NULL)
	{
		fprintf(stderr, "Error: File format error -- could not read file format (low/high/xtide)\n");
		goto ERROR;
	}

	
	if (strcmp(token, "low") == 0)
		format = low;
	else if (strcmp(token, "high") == 0)
		format = high;
	else if (strcmp(token, "xtide") == 0)
		format = xtide;
		
	//	Get units
	token = getnext(fs,false,kHeaderDelims);
	if (token == NULL)
	{
		fprintf(stderr, "Error: File format error -- could not read units (feet/meters)\n");
		goto ERROR;
	}

	if (strcmp(token, "meters") == 0)
		header.flags |= kMetersFlag;

	//	Get the tidesPerDay
	token = getnext(fs,false,kHeaderDelims);
	if (token == NULL)
	{
		fprintf(stderr, "Error: File format error -- could not read max tides per day\n");
		goto ERROR;
	}
	
	sscanf(token, "%d", &tidesPerDay);
	header.maxTidesPerDay = tidesPerDay;
	fwrite(&header, kTideDatabaseHeaderSize, 1, fd);
	
	switch(format)
	{
		case low:
		case high:
			err = convert_hltable(fs, fd, &header, format);
			break;
			
		case xtide:
			err = convert_xtide(fs, fd, &header, format);
			break;
	}
	
	fclose(fd);
	
ERROR:
	return;
}

int convert_hltable(FILE *fs, FILE *fd, TideDatabaseHeader *header, int format)
{
	TideEntry			tideEntry;
	char				*token;
	char				*ptide;
	char				*pfeet;
	char				mode;
	unsigned char		lastHour;
	int					count;
	int					tidecount;
	int					pm;
	int					day,month,year,hour,minute;
	float				height;

	memset(&tideEntry, 0, kTideEntrySize);
	count = 0;
	
	while ((token = getnext(fs, true, kFieldDelims)) != NULL)
	{
		if (sscanf(token,"%d/%d/%d",&month,&day,&year) != 3)
			goto TOKENERR;

		if (year < 100)
		{
			if (year < 90)
				year += 2000;
			else
				year += 1900;
		}
		
		if (count == 0)
		{
			header->stDay	= day;
			header->stMonth	= month;
			setshort(&(header->stYear),year);
		}

		mode = (format == low) ? kLowTide : kHighTide;
		tidecount = 0;
		pm = 0;
		lastHour = 0;
		while (tidecount < header->maxTidesPerDay)
		{
			
			ptide = getnext(fs, false, kFieldDelims);
			pfeet = getnext(fs, false, kFieldDelims);
			if (ptide && *ptide)
			{
				sscanf(ptide, "%d:%d", &hour, &minute);
				sscanf(pfeet, "%f", &height);
				// test for 12-hour mode
				if ((pm == false) && (hour == 12))
					hour = 0;
				if ((pm == false) && (hour < lastHour))
					pm = true;
				if ((pm == true) && (hour < 12))
					hour += 12;
				lastHour = hour;
				
				tideEntry.hour = hour;
				tideEntry.minute = minute;
				tideEntry.tideHeightHi = 0;
				if (height < 0)
				{
					tideEntry.tideHeightHi |= kTideNegativeFlag;
					height = - height;
				}
				tideEntry.tideHeightHi |= height;
				tideEntry.tideHeightLo = (height - tideEntry.tideHeightHi) * 100;
				if (height > 99)
				{
					fprintf(stderr,"ERROR: tide heights > 99 not supported\n");
					goto ERROR;
				}
					
				if (mode == kHighTide)
					tideEntry.hour |= kTideHighFlag;
				mode = !mode;
			}
			if (ptide == NULL)
				memset(&tideEntry, 0xFF, sizeof(TideEntry));
				
			fwrite(&tideEntry,kTideEntrySize,1,fd);
			tidecount ++;
		}
		count++;
		
	}
	header->endDay = day;
	header->endMonth = month;
	setshort(&(header->endYear), year);

	fseek(fd, kPDBFirstRecordOffset, SEEK_SET);
	fwrite(header, kTideDatabaseHeaderSize, 1, fd);
	
	return 0;

TOKENERR:
	fprintf(stderr, "Error: file format error\n");

ERROR:
	return -1;
}

int convert_xtide(FILE *fs, FILE *fd, TideDatabaseHeader *header, int format)
{
	TideEntry			tideEntry, tempEntry;
	char				*token;
	char				mode;
	int					lastDay;
	int					count;
	int					tidecount;
	int					pm;
	int					day,month,year,hour,minute;
	float				height;

	memset(&tideEntry, 0, kTideEntrySize);
	memset(&tempEntry, 0xFF, sizeof(TideEntry));
	
	tidecount = 0;
	count = 0;
	while ((token = getnext(fs, true, kFieldDelims)) != NULL)
	{
		mode = (strcmp(token,"high") == 0) ? kHighTide : kLowTide;
		
		token = getnext(fs, false, kFieldDelims);	// skip "tide"
		token = getnext(fs, false, kFieldDelims);	// skip DOW
		
		token = getnext(fs, false, kFieldDelims);	// get date
		if (token == NULL) goto TOKENERR;
		sscanf(token, "%d-%d-%d", &year, &month, &day);
		
		token = getnext(fs, false, kFieldDelims);	// get time
		if (token == NULL) goto TOKENERR;
		sscanf(token, "%d:%d", &hour, &minute);
		
		token = getnext(fs, false, kFieldDelims);	// get am/pm
		if (token == NULL) goto TOKENERR;
		pm = (strcmp(token,"pm") == 0) ? 1 : 0;
		
		token = getnext(fs, false, kFieldDelims);	//	skip timezone
		
		token = getnext(fs, false, kFieldDelims);	// get tide height
		if (token == NULL) goto TOKENERR;
		sscanf(token, "%f",&height);
		
		if (year < 100)
		{
			if (year < 90)
				year += 2000;
			else
				year += 1900;
		}
		
		if (count == 0)
		{
			lastDay = day;
			header->stDay	= day;
			header->stMonth	= month;
			setshort(&(header->stYear), year);
		}

		if ((pm == false) && (hour == 12))
			hour = 0;
		if ((pm == true) && (hour < 12))
			hour += 12;
		
		tideEntry.hour = hour;
		tideEntry.minute = minute;
		tideEntry.tideHeightHi = 0;
		if (height < 0)
		{
			tideEntry.tideHeightHi |= kTideNegativeFlag;
			height = -height;
		}
		tideEntry.tideHeightHi |= height;
		tideEntry.tideHeightLo = (height - tideEntry.tideHeightHi) * 100;
		if (height > 99)
		{
			fprintf(stderr,"ERROR: tide heights > 99 not supported\n");
			goto ERROR;
		}

		if (mode == kHighTide)
			tideEntry.hour |= kTideHighFlag;
		mode = !mode;

		// If we're on a new day, make sure we finish the last day
		if (lastDay != day)
		{
			while (tidecount < header->maxTidesPerDay)
			{
				fwrite(&tempEntry,kTideEntrySize,1,fd);
				tidecount ++;
			}
			tidecount = 0;
			lastDay = day;
		}
		fwrite(&tideEntry,kTideEntrySize,1,fd);
		tidecount ++;
		count++;
		
	}
	
	//	Make sure we clear out the rest of the array
	while (tidecount < header->maxTidesPerDay)
	{
		fwrite(&tempEntry,kTideEntrySize,1,fd);
		tidecount ++;
	}
	header->endDay = day;
	header->endMonth = month;
	setshort(&(header->endYear), year);

	fseek(fd, kPDBFirstRecordOffset, SEEK_SET);
	fwrite(header, kTideDatabaseHeaderSize, 1, fd);
	
	return 0;
	
TOKENERR:
	fprintf(stderr, "Error: file format error\n");
ERROR:
	return -1;
}

char *strlower(char *p)
{
	char	c;
	char	*s;
	if (p)
	{
		s = p;
		while ((c = *s) != 0)
			*s++ = tolower(c);
	}
	return(p);
}

void setlong(unsigned long *l, unsigned long value)
{
	unsigned char	*p;
	
	p = (unsigned char *) l;
	*p++ = value>>24;
	*p++ = value>>16;
	*p++ = value>>8;
	*p++ = value;
}

void setshort(unsigned short *s, unsigned short value)
{
	unsigned char	*p;
	
	p = (unsigned char *) s;
	*p++ = value>>8;
	*p++ = value;
}
